# -*- coding: utf-8 -*-

import sys
import stdlib_list
import warnings


def pystdlib():
    """Return a set of all module-names in the Python standard library.
    """
    # stdlib_list has been transferred to the pypi team 
    # (https://github.com/pypi/stdlib-list) and now works for Python 3.10+
    # Removing the workaround since it doesn't contain all the symbols that
    # stdlib_list does.
    #
    # if sys.version_info[:2] >= (3, 10):
    #     # Python 3.10 has this functionality built-in.
    #     return list(sys.stdlib_module_names | set(sys.builtin_module_names))

    curver = '.'.join(str(x) for x in sys.version_info[:2])
    if curver not in stdlib_list.short_versions:
        # if stdlib_list doesn't know about our version, then use the last
        # version that stdlib_list knows about (not perfect, but it will
        # allow downstream packages to test on pre-release versions of
        # Python - which are difficult for stdlib_list to support).
        warnings.warn(
            ("stdlib_list does't support Python %s yet, "
             "pydeps will use symbols from %s for now.") % (
                 curver, stdlib_list.long_versions[-1]
             )
        )
        curver = stdlib_list.long_versions[-1]
    return (set(stdlib_list.stdlib_list(curver)) | {
        '_LWPCookieJar', '_MozillaCookieJar', '_abcoll', 'email._parseaddr',
        'email.base64mime', 'email.feedparser', 'email.quoprimime',
        'encodings', 'genericpath', 'ntpath', 'nturl2path', 'os2emxpath',
        'posixpath', 'sre_compile', 'sre_parse', 'unittest.case',
        'unittest.loader', 'unittest.main', 'unittest.result',
        'unittest.runner', 'unittest.signals', 'unittest.suite',
        'unittest.util', '_threading_local', 'sre_constants', 'strop',
        'repr', 'opcode', 'nt', 'encodings.aliases',
        '_bisect', '_codecs', '_collections', '_functools', '_hashlib',
        '_heapq', '_io', '_locale', '_LWPCookieJar', '_md5',
        '_MozillaCookieJar', '_random', '_sha', '_sha256', '_sha512',
        '_socket', '_sre', '_ssl', '_struct', '_subprocess',
        '_threading_local', '_warnings', '_weakref', '_weakrefset',
        '_winreg'
    }) - {'__main__'}



# -*- coding: utf-8 -*-
from __future__ import print_function
import os
import site


def _find_top_level_file(site_pkg_dir, pth):
    if pth.endswith('.dist-info') or pth.endswith('.egg-info'):
        top_level_fname = os.path.join(site_pkg_dir, pth, 'top_level.txt')
    elif pth.endswith('.egg'):
        top_level_fname = os.path.join(site_pkg_dir, pth, 'EGG-INFO', 'top_level.txt')
    else:
        top_level_fname = None
    return top_level_fname


def _extract_pkg_name(pth):
    name = pth.rsplit('.', 1)[0]
    name_no_version = name.split('-', 1)[0]
    return name_no_version.replace('_', '-')


def find_package_names():
    # initialize with well-known packages that don't seem to have a top_level.txt
    res = {
        'yaml': 'PyYAML',
        'Crypto': 'pycrypto',
    }
    site_package_dirs = [site.getusersitepackages()]
    site_package_dirs += site.getsitepackages()

    for site_packages in reversed(site_package_dirs):
        if not os.path.isdir(site_packages):
            continue

        for pth in os.listdir(site_packages):
            top_level_fname = _find_top_level_file(site_packages, pth)
            if top_level_fname is None:
                continue

            pkgname = _extract_pkg_name(pth)

            if not os.path.exists(top_level_fname):
                if pkgname not in res.values():
                    print("ERR:", pth, 'has not top_level.txt')
                continue

            with open(top_level_fname) as fp:
                modnames = fp.read().split()

            for modname in modnames:
                modname = modname.replace('/', '.')
                if modname.startswith(r'win32\lib'):
                    modname = modname.rsplit('\\')[1]
                res[modname] = pkgname

    return res


if __name__ == "__main__":
    import json
    print(json.dumps(find_package_names(), indent=4, sort_keys=True))



# -*- coding: utf-8 -*-
"""Color calculations.
"""
import colorsys

# noinspection PyAugmentAssignment
# import hashlib

START_COLOR = 0  # Value can be changed from command-line argument


def frange(start, end, step):
    """Like range(), but with floats.
    """
    val = start
    while val < end:
        yield val
        val += step


def distinct_hues(count):
    """Return ``count`` hues, equidistantly spaced.
    """
    for i in frange(0., 360., 360. / count):
        hue = ((i + START_COLOR) % 360) / 360.
        yield hue


class ColorSpace(object):
    def __init__(self, nodes):
        self.nodes = {}
        for node in nodes:
            parts = node.name.split('.')
            self.add_to_tree(parts, self.nodes)
        self.basecolors = distinct_hues(len(self.nodes))
        self.colors = dict(zip(sorted(self.nodes.keys()), self.basecolors))

    def add_to_tree(self, parts, tree):
        if not parts:
            return
        first, rest = parts[0], parts[1:]
        if first not in tree:
            tree[first] = {}
        self.add_to_tree(rest, tree[first])

    def color(self, src):
        nodename = src.name
        parts = nodename.split('.')
        hue = self.colors[parts[0]]
        saturation = min(0.95, 0.4 + 0.1 * (src.out_degree - 1))
        lightness = max(0.3, 0.5 - 0.02 * (src.in_degree - 1))
        bg = rgb2eightbit(colorsys.hls_to_rgb(hue, lightness, saturation))
        black = (0, 0, 0)
        white = (255, 255, 255)
        fg = foreground(bg, black, white)
        return bg, fg

    def __str__(self):  # pragma: nocover
        import pprint
        return pprint.pformat(self.colors)


def rgb2eightbit(rgb):
    """Convert floats in [0..1] to integers in [0..256)
    """
    return tuple(int(x * 256) for x in rgb)


def name2rgb(hue):
    """Originally used to calculate color based on module name.
    """
    r, g, b = colorsys.hsv_to_rgb(hue / 360.0, .8, .7)
    return tuple(int(x * 256) for x in [r, g, b])


def brightness(r, g, b):
    """From w3c (range 0..255).
    """
    return (r * 299 + g * 587 + b * 114) / 1000


def brightnessdiff(a, b):
    """greater than 125 is good.
    """
    return abs(brightness(*a) - brightness(*b))


def colordiff(rgb1, rgb2):
    """From w3c (greater than 500 is good).
       (range [0..765])
    """
    (r, g, b) = rgb1
    (r2, g2, b2) = rgb2
    return (
        max(r, r2) - min(r, r2) +
        max(g, g2) - min(g, g2) +
        max(b, b2) - min(b, b2)
    )


def foreground(background, *options):
    """Find the best foreground color from `options` based on `background`
       color.
    """
    def absdiff(a, b):
        return brightnessdiff(a, b)
        # return 3 * brightnessdiff(a, b) + colordiff(a, b)
    diffs = [(absdiff(background, color), color) for color in options]
    diffs.sort(reverse=True)
    return diffs[0][1]


def rgb2css(rgb):
    """Convert rgb to hex.
    """
    return '#%02x%02x%02x' % rgb



# -*- coding: utf-8 -*-
from __future__ import print_function
from collections import defaultdict
import fnmatch
from .pycompat import zip_longest
import json
import os
import pprint
import re
import enum

from . import colors, cli
import sys
import logging
log = logging.getLogger(__name__)

# we're normally not interested in imports of std python packages.
PYLIB_PATH = {
    # in virtualenvs that see the system libs, these will be different.
    os.path.split(os.path.split(pprint.__file__)[0])[0].lower(),
    os.path.split(os.__file__)[0].lower()
}


class imp(enum.Enum):
    C_BUILTIN = 6
    C_EXTENSION = 3
    IMP_HOOK = 9
    PKG_DIRECTORY = 5
    PY_CODERESOURCE = 8
    PY_COMPILED = 2
    PY_FROZEN = 7
    PY_RESOURCE = 4
    PY_SOURCE = 1
    UNKNOWN = 0


class Source(object):
    """A node (contained) in the dependency graph.

       It contains info about which modules are imported by this source,
       and which modules import this source.
    """
    def __init__(self, name, path=None, imports=(), exclude=False, args=None):
        self.args = args or {}
        self.name = name
        # self.kind = kind
        self.path = path             # needed here..?
        self.imports = set(imports)  # modules we import
        self.imported_by = set()     # modules that import us
        self.bacon = sys.maxsize      # bacon distance
        self.excluded = exclude

    @property
    def name_parts(self):
        return self.name.split('.')

    @property
    def module_depth(self):
        return self.name.count('.')

    @property
    def path_parts(self):
        p = self.path or ""
        return p.replace('\\', '/').lower().split('/')

    @property
    def in_degree(self):
        """Number of incoming arrows.
        """
        return len(self.imports)

    @property
    def out_degree(self):
        """Number of outgoing arrows.
        """
        return len(self.imported_by)

    @property
    def degree(self):
        return self.in_degree + self.out_degree

    def is_noise(self):
        """Is this module just noise?  (too common either at top or bottom of
           the graph).
        """
        noise = self.args['noise_level']
        if not (self.in_degree and self.out_degree):
            return self.degree > noise
        return False

    def __json__(self):
        res = dict(
            name=self.name,
            path=self.path,
            # kind=str(self.kind),
            bacon=self.bacon,
        )
        if self.excluded:
            res['excluded'] = 'EXCLUDED'
        if self.imports:
            res['imports'] = list(sorted(self.imports))
        if self.imported_by:
            res['imported_by'] = list(sorted(self.imported_by))
        return res

    def __str__(self):
        return "%s (%s)" % (self.name, self.path)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.name == other.name

    def __lt__(self, other):
        return self.name < other.name

    def __le__(self, other):
        return self.name <= other.name

    def __gt__(self, other):
        return self.name > other.name

    def __ge__(self, other):
        return self.name >= other.name

    def __repr__(self):
        return json.dumps(self.__json__(), indent=4)

    def __iadd__(self, other):
        """Merge other into self.
        """
        if self.name == other.name and self.imports == other.imports and self.bacon == other.bacon:
            return self
        log.debug("iadd lhs: %r", self)
        log.debug("iadd rhs: %r", other)
        assert self.name == other.name
        self.path = self.path or other.path
        self.imports |= other.imports
        self.imported_by |= other.imported_by
        self.bacon = min(self.bacon, other.bacon)
        self.excluded = self.excluded or other.excluded
        log.debug("iadd result: %r", self)
        return self

    def get_label(self, splitlength=0, rmprefix=None):
        name = self.name
        if rmprefix:
            for prefix in rmprefix:
                if self.name.startswith(prefix):
                    name = self.name[len(prefix):]
                    break
        if splitlength and len(name) > splitlength and '.' in name:
            return '\\.\\n'.join(name.split('.'))  # pragma: nocover
        return name

    @property
    def label(self):
        """Convert a module name to a formatted node label. This is a default
           policy - please override.
        """
        return self.get_label(splitlength=14)


class DepGraph(object):
    """The dependency graph.

       It is the output of :func:`pydeps.py2depgraph.py2dep`
    """
    skip_modules = """
        os sys qt time __future__ types re string bdb pdb __main__
        south
        """.split()

    def __init__(self, depgraf, types, target, **args):
        # depgraph is py2depgraph.MyModulefinder._depgraph
        log.debug("DepGraph: depgraf=%r", depgraf)

        self.curhue = 150  # start with a green-ish color
        self.colors = {}
        self.cycles = []
        self.cyclenodes = set()
        self.cyclerelations = set()
        self.max_module_depth = args.get('max_module_depth', 0)
        self.target = target

        self.args = args

        #: dict[module_name] -> Source object
        self.sources = {}
        self.skiplist = [re.compile(fnmatch.translate(arg)) for arg in args['exclude']]
        self.skiplist += [re.compile('^%s$' % fnmatch.translate(arg)) for arg in args['exclude_exact']]
        # depgraf = {name: imports for (name, imports) in depgraf.items()}

        for name, imports in depgraf.items():
            log.debug("depgraph name=%r imports=%r", name, imports)
            src = Source(
                name=self.source_name(name),
                imports=[self.source_name(n) for n in imports.keys()],  # values handled below
                args=args,
                exclude=self._exclude(name),
            )
            log.debug("depgraph src=%r", src)
            self.add_source(src)
            for iname, path in imports.items():
                src = Source(
                    name=self.source_name(iname, path),
                    path=path,
                    args=args,
                    exclude=self._exclude(iname)
                )
                self.add_source(src)

        self.module_count = len(self.sources)
        cli.verbose(1, "there are", self.module_count, "total modules")

        self.connect_generations()
        if self.args['show_cycles']:
            self.find_import_cycles()
        self.calculate_bacon()
        if self.args['show_raw_deps']:
            print(self)

        self.exclude_noise()
        self.exclude_bacon(self.args['max_bacon'])
        self.only_filter(self.args.get('only'))

        excluded = [v for v in list(self.sources.values()) if v.excluded]
        # print "EXCLUDED:", excluded
        self.skip_count = len(excluded)
        cli.verbose(1, "skipping", self.skip_count, "modules")
        for module in excluded:
            # print 'exclude:', module.name
            cli.verbose(2, "  ", module.name)

        self.remove_excluded()

        if not self.args['show_deps']:
            cli.verbose(3, self)

    def source_name(self, name, path=None):
        """Returns the module name, possibly limited by --max-module-depth.
        """
        res = name
        if name == "__main__" and self.target.is_pysource:
            # use the target file name directly if we're working on a 
            # single file
            return self.target.fname

        if name == "__main__" and path:
            # use the path to the main module if we're working on a module.
            res = path.replace('\\', '/').replace('/', '.')
            if self.args.get('verbose', 0) >= 2:  # pragma: nocover
                print("changing __main__ =>", res)

        if self.max_module_depth > 0:
            res = '.'.join(res.split('.')[:self.max_module_depth])
        return res

    def __json__(self):
        return json.dumps(self.sources, indent=4, sort_keys=True,
                          default=lambda obj: obj.__json__() if hasattr(obj, '__json__') else obj)

    def levelcounts(self):
        pass

    def get_colors(self, src, colorspace=None):
        if colorspace is None:
            if src.basename not in self.colors:
                h = self.curhue
                # self.curhue += 7  # relative prime with 360
                self.curhue += 37  # relative prime with 360
                self.curhue %= 360
                # print "NAME:", src.name, "BASENAME:", src.basename
                bg = colors.name2rgb(h)
                black = (0, 0, 0)
                white = (255, 255, 255)
                fg = colors.foreground(bg, black, white)
                self.colors[src.basename] = bg, fg
            return self.colors[src.basename]
        else:
            return colorspace.color(src)

    def _is_pylib(self, path):
        log.info('path %r in PYLIB_PATH %r => %s', path, PYLIB_PATH, path in PYLIB_PATH)
        return path in PYLIB_PATH

    def proximity_metric(self, a, b):
        """Return the weight of the dependency from a to b. Higher weights
           usually have shorter straighter edges. Return 1 if it has normal
           weight. A value of 4 is usually good for ensuring that a related
           pair of modules are drawn next to each other.

           Returns an int between 1 (unknown, default), and 4 (very related).
        """
        res = 0
        for ap, bp, n in zip(a.name_parts, b.name_parts, list(range(4))):
            res += ap == bp
            if n >= 3:
                break
        if res == 0: res = 1    # noqa
        return 4 if res > 4 else res

    def dissimilarity_metric(self, a, b):
        """Return non-zero if references to this module are strange, and
           should be drawn extra-long. The value defines the length, in
           rank. This is also good for putting some vertical space between
           seperate subsystems.

           Returns an int between 1 (default) and 4 (highly unrelated).
        """
        res = 4
        for an, bn, n in zip_longest(a.name_parts, b.name_parts, list(range(4))):
            res -= an == bn
            if n >= 3:
                break
        return 4 if res > 4 else res

    def _exclude(self, name):
        return any(skip.match(name) for skip in self.skiplist)

    def add_source(self, src):
        if src.name in self.sources:
            log.debug("ADD-SOURCE[+=]\n%r", src)
            self.sources[src.name] += src   # merge
        else:
            log.debug("ADD-SOURCE[=]\n%r", src)
            self.sources[src.name] = src

    def __getitem__(self, item):
        return self.sources[item]

    def __iter__(self):
        visited = set(self.skip_modules) | set(self.args['exclude'])

        def visit(src):
            if src.name in visited:
                return
            visited.add(src.name)
            for name in src.imports:
                impmod = self.sources[name]

                # FIXME: why do we want to exclude **/*/__init__.py? This line
                # causes `collections` package in py3 to be excluded.
                # if impmod.path and not impmod.path.endswith('__init__.py'):
                if not src.name.startswith(impmod.name + "."):
                    yield impmod, src
                visit(impmod)

        for _src in self.sources.values():
            for source in visit(_src):
                cli.verbose(4, "Yielding", source[0], source[1])
                yield source

    def __repr__(self):
        return json.dumps(self.sources, indent=4, sort_keys=True,
                          default=lambda obj: obj.__json__() if hasattr(obj, '__json__') else obj)

    def find_import_cycles(self):
        def traverse(node, path):
            if node.name in self.cyclenodes:
                return

            if node.name in path:
                # found cycle
                cycle = path[path.index(node.name):] + [node.name]
                self.cycles.append(cycle)
                for nodename in cycle:
                    self.cyclenodes.add(nodename)
                for i in range(len(cycle) - 1):
                    self.cyclerelations.add(
                        (cycle[i], cycle[i + 1])
                    )
                return

            for impmod in node.imports:
                traverse(self.sources[impmod], path + [node.name])

        for src in list(self.sources.values()):
            traverse(src, [])

    def connect_generations(self):
        """Traverse depth-first adding imported_by.
        """
        for src in self.sources.values():
            for _child in src.imports:
                if _child in self.sources:
                    child = self.sources[_child]
                    child.imported_by.add(src.name)

    def calculate_bacon(self):
        count = defaultdict(int)

        def bacon(src, n):
            count[src.name] += 1
            if src.bacon <= n:
                return
            src.bacon = min(src.bacon, n)
            for imp in src.imports:
                bacon(self.sources[imp], n + 1)

        if '__main__' in self.sources:
            bacon(self.sources['__main__'], 0)
        elif self.args['dummyname'] in self.sources:
            bacon(self.sources[self.args['dummyname']], 0)

    def exclude_noise(self):
        for src in list(self.sources.values()):
            if src.excluded:
                continue
            if src.is_noise():
                cli.verbose(2, "excluding", src, "because it is noisy:", src.degree)
                src.excluded = True
                self._add_skip(src.name)

    def exclude_bacon(self, limit):
        """Exclude modules that are more than `limit` hops away from __main__.
        """
        for src in list(self.sources.values()):
            if src.bacon > limit:
                src.excluded = True
                self._add_skip(src.name)

    def only_filter(self, paths):
        """Exclude nodes that have a prefix in paths.
        """
        if not paths:
            return
        paths = set(paths)

        def should_include(node):
            for p in paths:
                if node.name.startswith(p):
                    return True
            return False

        for src in list(self.sources.values()):
            if not should_include(src):
                src.excluded = True
                # print "Excluding bacon:", src.name
                self._add_skip(src.name)

    def remove_excluded(self):
        """Remove all sources marked as excluded.
        """
        sources = list(self.sources.values())
        for src in sources:
            if src.excluded:
                del self.sources[src.name]
            src.imports = [m for m in src.imports if not self._exclude(m)]
            src.imported_by = [m for m in src.imported_by if not self._exclude(m)]

    def _add_skip(self, name):
        # print 'add skip:', name
        self.skiplist.append(re.compile(fnmatch.translate(name)))



"""
Python's modulefinder._find_module has a bug that breaks a number of popular
packages.

This is a copy of the standard lib's imp._find_module which sort of does the
right thing, ie. ignores namespace packages instead of crashing.

This is vendorized/copied here to prevent the warning error that the regular 
imp module causes.
"""

from _imp import is_builtin, is_frozen
from importlib import machinery
import sys
import os
import tokenize


PY_SOURCE = 1
PY_COMPILED = 2
C_EXTENSION = 3
PKG_DIRECTORY = 5
C_BUILTIN = 6
PY_FROZEN = 7


def _get_suffixes():
    extensions = [(s, 'rb', C_EXTENSION) for s in machinery.EXTENSION_SUFFIXES]
    source = [(s, 'r', PY_SOURCE) for s in machinery.SOURCE_SUFFIXES]
    bytecode = [(s, 'rb', PY_COMPILED) for s in machinery.BYTECODE_SUFFIXES]

    return extensions + source + bytecode


def find_module(name, path=None):
    if not isinstance(name, str):
        raise TypeError("'name' must be a str, not {}".format(type(name)))
    elif not isinstance(path, (type(None), list)):
        # Backwards-compatibility
        raise RuntimeError("'path' must be None or a list, "
                           "not {}".format(type(path)))

    if path is None:
        if is_builtin(name):
            return None, None, ('', '', C_BUILTIN)
        elif is_frozen(name):
            return None, None, ('', '', PY_FROZEN)
        else:
            path = sys.path

    for entry in path:
        package_directory = os.path.join(entry, name)
        for suffix in ['.py', machinery.BYTECODE_SUFFIXES[0]]:
            package_file_name = '__init__' + suffix
            file_path = os.path.join(package_directory, package_file_name)
            if os.path.isfile(file_path):
                return None, package_directory, ('', '', PKG_DIRECTORY)
        for suffix, mode, type_ in _get_suffixes():
            file_name = name + suffix
            file_path = os.path.join(entry, file_name)
            if os.path.isfile(file_path):
                break
        else:
            continue
        break  # Break out of outer loop when breaking out of inner loop.
    else:
        raise ImportError('No module named {!r}'.format(name), name=name)

    encoding = None
    if 'b' not in mode:
        with open(file_path, 'rb') as file:
            encoding = tokenize.detect_encoding(file.readline)[0]
    file = open(file_path, mode, encoding=encoding)
    return file, file_path, (suffix, mode, type_)



from __future__ import print_function, unicode_literals
from io import StringIO
import textwrap
import json
import argparse
# from devtools import debug
from .configs import Config, typefns, identity



DEFAULT_NONE = '____'




class Argument(object):
    def __init__(self, *flags, **args):
        if 'choices' in args and args['choices'] is None:
            del args['choices']
        if 'container' in args:
            del args['container']
        self._args = args
        self._flags = flags
        # self.__dict__.update(args)

    def __json__(self):
        return self.__dict__

    def typename(self):
        if 'kind' in self._args:
            return self._args['kind']
        if self._args.get('action') in {'store_true', 'store_false'}:
            return 'BOOL'
        tp = self._args.get('type')
        if tp is not None:
            return tp.__name__.upper()
        v = self._args.get('default')
        if v is None:
            return DEFAULT_NONE
        return v.__class__.__name__.upper()
    
    def typefn(self):
        return typefns.get(self.typename(), identity)

    def pytype(self):
        argname = self.argname()
        if argname == 'fname':
            return 'str'
        if 'choices' in self._args:
            return 'Literal[%s]' % ', '.join(repr(x) for x in self._args['choices'])
        if 'type' in self._args:
            res = self._args['type']
        elif self._args.get('action') in {'store_true', 'store_false'}:
            res = bool
        elif self._args.get('kind', 'unknown') is None:
            return 'None'
        elif self._args.get('kind', '').startswith("FNAME"):
            res = str
        else:
            res = self._args.get('default').__class__
        if res is None:
            return 'None'
        return res.__name__

    def argname(self):
        if 'dest' in self._args:
            return self._args['dest']
        return self._flags[0].lstrip('-').replace('-', '_')
        # return self._flags[-1].lstrip('-').replace('-', '_')
    
    def help(self):
        if 'help' in self._args:
            return self._args['help']
        return ''

    def default(self):
        if 'default' in self._args:
            return self._args['default']
        if self._args.get('action') == 'store_true':
            return False
        if self._args.get('action') == 'store_false':
            return True
        return DEFAULT_NONE

    def add_to_parser(self, parser):
        args = self._args
        args.pop('default', None)  # remove default value
        args.pop('kind', None)     # remove our attributes
        if args.get('action') in {'store_true', 'store_false'}:
            args['default'] = None
        # debug("ADD_TO_PARSER:", self._flags, args)
        parser.add_argument(*self._flags, **args)


class Namespace(object):
    def __init__(self, ns):
        self.ns = ns

    def __repr__(self):
        return json.dumps(vars(self.ns), indent=4, sort_keys=True)

    def items(self):
        return list(dict(vars(self.ns)).items())

    def __getitem__(self, key):
        return getattr(self.ns, key)

    def __setitem__(self, key, val):
        setattr(self.ns, key, val)

    def __delitem__(self, key):
        delattr(self.ns, key)

    def __getattr__(self, key):
        return getattr(self.ns, key)


class Arguments(object):
    def __init__(self, config_files=None, debug=False, *posargs, parents=None, **kwargs):
        if config_files is None:
            config_files = []
        # passthrough to argparse
        self.posargs = posargs
        self.kwargs = kwargs
        # print("CONFIG_FILES:", config_files)

        self.debug = debug
        self.arglist = []
        self.args = {}
        self.config_files = config_files
        self.argtypes = {}
        self.defaults = {}
        self.parents = parents

    def load_config_files(self):
        for filename in self.config_files:
            if filename == '.pydeps' or filename.endswith('.pydeps'):
                self.load_pydeps_config(filename)

    def write_default_config(self):
        """Utility function to create .configs.Config

           # XXX: a more general utility to create configs from argparse
           #      would be nice...
        """
        fp = StringIO()
        # fp = sys.stdout
        print("class Config(object):", file=fp)

        arglist = self.arglist

        if self.parents:
            parent_actions = []
            for parent in self.parents:
                for action in parent._actions:
                    # if isinstance(action, argparse._StoreAction):
                    parent_actions.append(Argument(
                        *action.option_strings, **action.__dict__
                    ))
            arglist = parent_actions + arglist

        for arg in arglist:
            debug(arg._args)
            
            default = arg.default()
            if default == DEFAULT_NONE:
                default = None
            elif isinstance(default, str):
                default = repr(default)

            help = textwrap.wrap(arg.help(), 80 - 7)  # 7 = len("    #: ")
            for helpline in help:
                print("", file=fp)
                print("    #: {help}".format(help=helpline), end="", file=fp)
            print("", file=fp)

            typename = arg.pytype()
            if typename == 'list' and default is None:
                typename = 'Optional[List[str]]'
            elif default is None:
                typename = f'Optional[{typename}]'
            elif default == [] and typename == 'list':
                typename = 'List[str]'
            # add py3 type annotations
            # print(f"    {arg.argname()}: {typename} = {default}", file=fp)
            
            # witout type annotations
            print("    {argname} = {default}".format(argname=arg.argname(), default=default), file=fp)

        print("", file=fp)
        print("    def set_field(self, field, value):", file=fp)
        for arg in self.arglist:
            print("        if field == '{argname}':".format(argname=arg.argname()), file=fp)
            print("            self.{argname} = {typefn}(value)".format(argname=arg.argname(), typefn=arg.typefn().__name__), file=fp)
        print("", file=fp)
        # debug(res)
        res = fp.getvalue()
        print(res)
        return res

    def parse_args(self, argv):
        self.kwargs['parents'] = self.parents
        p = argparse.ArgumentParser(*self.posargs, **self.kwargs)
        for arg in self.arglist:
            arg.add_to_parser(p)
        args = Namespace(p.parse_args(argv))

        config = Config.load(self.config_files)
        config.update({k : v for k, v in args.items() if v is not None})

        # print('---- yaml ---------------')
        # print(config.write_yaml())
        # print('---- json ---------------')
        # print(config.write_json())
        # print('---- ini ---------------')
        # print(config.write_ini())
        # print('---- toml ---------------')
        # print(config.write_toml())
        # print('-------------------')

        return config

    def add(self, *flags, **kwargs):
        # debug(flags, kwargs)
        arg = Argument(*flags, **kwargs)
        self.arglist.append(arg)
        argname = arg.argname()
        self.args[argname] = arg
        self.argtypes[argname] = arg.typename()
        self.defaults[argname] = arg.default()

    def __repr__(self):
        return json.dumps(dict(
            types=self.argtypes,
            defaults=self.defaults,
        ), indent=4, sort_keys=True)



# Based on original code, Copyright 2004 Toby Dickenson,
# with changes 2014 (c) Bjorn Pettersen
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject
# to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from .render_context import RenderBuffer
from . import colors


class PyDepGraphDot(object):
    def __init__(self, **kw):
        self.kw = kw

    def render(self, depgraph, ctx):
        with ctx.graph():
            visited = set()
            drawn = set()

            for aname, bname in depgraph.cyclerelations:
                try:
                    a = depgraph.sources[aname]
                    b = depgraph.sources[bname]
                except KeyError:
                    continue
                drawn.add((bname, aname))
                ctx.write_rule(
                    aname, bname,
                    weight=depgraph.proximity_metric(a, b),
                    minlen=depgraph.dissimilarity_metric(a, b),
                )

            for a, b in sorted(depgraph):
                # b imports a
                aname = a.name
                bname = b.name
                if (bname, aname) in drawn:
                    continue
                drawn.add((bname, aname))

                ctx.write_rule(
                    aname, bname,
                    weight=depgraph.proximity_metric(a, b),
                    minlen=depgraph.dissimilarity_metric(a, b))

                visited.add(a)
                visited.add(b)

            space = colors.ColorSpace(visited)
            for src in sorted(visited):
                bg, fg = depgraph.get_colors(src, space)
                kwargs = {}

                if src.name in depgraph.cyclenodes:
                    kwargs['shape'] = 'octagon'

                ctx.write_node(
                    src.name,
                    label=src.get_label(splitlength=14,
                                        rmprefix=self.kw.get('rmprefix')),
                    fillcolor=colors.rgb2css(bg),
                    fontcolor=colors.rgb2css(fg),
                    **kwargs
                )

        return ctx.text()


class CycleGraphDot(object):
    def __init__(self, **kw):
        self.kw = kw

    def render(self, depgraph, ctx):
        with ctx.graph(concentrate=False):
            visited = set()
            drawn = set()

            for aname, bname in depgraph.cyclerelations:
                try:
                    a = depgraph.sources[aname]
                    b = depgraph.sources[bname]
                except KeyError:
                    continue
                drawn.add((bname, aname))
                ctx.write_rule(
                    bname, aname,
                    weight=depgraph.proximity_metric(a, b),
                    minlen=depgraph.dissimilarity_metric(a, b),
                )
                visited.add(a)
                visited.add(b)

            space = colors.ColorSpace(visited)
            for src in visited:
                bg, fg = depgraph.get_colors(src, space)
                kwargs = {}

                if src.name in depgraph.cyclenodes:
                    kwargs['shape'] = 'octagon'

                ctx.write_node(
                    src.name, label=src.label,
                    fillcolor=colors.rgb2css(bg),
                    fontcolor=colors.rgb2css(fg),
                    **kwargs
                )

        return ctx.text()


def dep2dot(target, depgraph, **kw):
    dotter = PyDepGraphDot(**kw)
    ctx = RenderBuffer(target, **kw)
    return dotter.render(depgraph, ctx)


def cycles2dot(target, depgraph, **kw):
    dotter = CycleGraphDot(**kw)
    ctx = RenderBuffer(target, **kw)
    return dotter.render(depgraph, ctx)



# -*- coding: utf-8 -*-
import os
import sys

import pydeps.cli
from pydeps import pydeps
from pydeps.target import Target
from tests.filemaker import create_files
from tests.simpledeps import simpledeps


def test_dep2dot():
    files = """
        foo:
            - __init__.py
            - a.py: |
                from . import b
            - b.py
    """
    with create_files(files) as workdir:
        assert simpledeps('foo', '-LDEBUG -vv') == {
            'foo.b -> foo.a'
        }

        args = pydeps.cli.parse_args(["foo", "--noshow"])
        pydeps.pydeps(**args)
        assert os.path.exists(os.path.join(workdir, 'foo.svg'))



# -*- coding: utf-8 -*-
from pydeps.dot import dot, cmd2args

def test_svg(tmpdir):
    tmpdir.chdir()
    ab = tmpdir.join('ab.svg')
    dot(u"""
    digraph G {
        a -> b
    }
    """, o=ab.basename)
    assert ab.exists()


def test_svg_str(tmpdir):
    tmpdir.chdir()
    ab = tmpdir.join('ab.svg')
    dot("""
    digraph G {
        a -> b
    }
    """, o=ab.basename)
    assert ab.exists()


def test_boolopt(tmpdir):
    tmpdir.chdir()
    ab = tmpdir.join('ab.svg')
    dot("""
    digraph G {
        a -> b
    }
    """, x=True, o=ab.basename)
    print(tmpdir.listdir())
    assert ab.exists()


def test_obj(tmpdir):  # pragma: nocover
    GRAPH = u"""
    digraph G {
        a -> b
    }
    """
    import sys
    if sys.version_info >= (3,):
        class MyClass:
            def __str__(self):
                return GRAPH
    else:
        class MyClass(object):
            def __unicode__(self):
                return GRAPH

    tmpdir.chdir()
    ab = tmpdir.join('ab.svg')
    dot(MyClass(), x=True, o=ab.basename)
    print(tmpdir.listdir())
    assert ab.exists()


def test_cmd2args():
    assert cmd2args([1, 2]) == [1, 2]



# -*- coding: utf-8 -*-
from tests.filemaker import create_files
from tests.simpledeps import simpledeps


def test_cycle():
    files = """
        relimp:
            - __init__.py
            - a.py: |
                from . import b
            - b.py: |
                from . import a
    """
    with create_files(files, cleanup=False) as workdir:
        print("WORKDIR:", workdir)
        deps = simpledeps('relimp')
        assert 'relimp.a -> relimp.b' in deps
        assert 'relimp.b -> relimp.a' in deps



# -*- coding: utf-8 -*-
import ast

from pydeps.pydeps import pydeps
from tests.filemaker import create_files
from tests.simpledeps import simpledeps


def test_relative_imports(capsys):
    files = """
        foo:
            - __init__.py
            - a.py: |
                from bar import b
        bar:
            - __init__.py
            - b.py
    """
    with create_files(files) as workdir:
        assert simpledeps('foo') == {
            'bar -> foo.a',
            'bar.b -> foo.a'
        }
        pydeps(fname='foo', externals=True)
        io = capsys.readouterr()
        assert ast.literal_eval(io.out) == ['bar']



# -*- coding: utf-8 -*-
import sys
from tests.filemaker import create_files
from tests.simpledeps import simpledeps
import pytest


@pytest.mark.skipif(sys.platform != 'win32', reason=".pyw files only exist on windows")
def test_from_pyw():
    files = """
        baz.pyw: |
            import foo.a
            import bar
        foo:
            - __init__.py
            - a.py: |
                from bar import py
        bar:
            - __init__.py
    """
    with create_files(files) as workdir:
        assert simpledeps('baz.pyw', '--show-deps -LINFO -vv') == {
            'foo.a -> baz.pyw',
            'foo -> baz.pyw',
            'bar -> baz.pyw',
            'bar -> foo.a',
        }



# -*- coding: utf-8 -*-
import os
# from devtools import debug
from pydeps.pydeps import call_pydeps
from tests.filemaker import create_files
from tests.simpledeps import simpledeps


def test_py2depgraph(capsys):
    files = """
        - a.py: |
            import b
        - b.py
    """
    with create_files(files) as workdir:
        assert simpledeps('a.py') == {'b -> a.py'}


def test_pydeps_api():
    files = """
        - a.py: |
            import b
        - b.py
    """
    with create_files(files) as workdir:
        call_pydeps('a.py', no_show=True, show_dot=True, dot_out='a.dot')
        # debug(os.getcwd())
        # debug(os.listdir('.'))
        with open('a.dot') as f:
            dot = f.read()
            # debug(dot)
        assert 'b -> a_py' in dot



# Copyright 2004,2009 Toby Dickenson
# Changes 2014 (c) Bjorn Pettersen
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject
# to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import print_function

import json
import os
import sys
from collections import defaultdict

import enum

from .dummymodule import DummyModule
from .pystdlib import pystdlib
from . import depgraph
from . import mf27
import logging
log = logging.getLogger(__name__)

PYLIB_PATH = depgraph.PYLIB_PATH


class imp(enum.IntEnum):
    C_BUILTIN = 6
    C_EXTENSION = 3
    IMP_HOOK = 9
    PKG_DIRECTORY = 5
    PY_CODERESOURCE = 8
    PY_COMPILED = 2
    PY_FROZEN = 7
    PY_RESOURCE = 4
    PY_SOURCE = 1


class Module(object):
    def __init__(self, name, file=None, path=None):
        self.__name__ = name
        self.__file__ = file
        self.__path__ = path
        self.__code__ = None
        # The set of global names that are assigned to in the module.
        # This includes those names imported through starimports of
        # Python modules.
        self.globalnames = {}
        # The set of starimports this module did that could not be
        # resolved, ie. a starimport from a non-Python module.
        self.starimports = {}

    @property
    def shortname(self):
        if self.__name__ == "__main__":
            return self.__file__[:-3].replace('\\', '.')
        return self.__name__

    def __repr__(self):
        return "Module(name=%s, file=%r, path=%r)" % (
            self.__name__,
            (self.__file__ or "").replace('\\', '/'),
            self.__path__
        )


class MyModuleFinder(mf27.ModuleFinder):
    def __init__(self, syspath, *args, **kwargs):
        self.args = kwargs
        self.verbose = kwargs.get('verbose', 0)

        # include all of python std lib (incl. C modules)
        self.include_pylib_all = kwargs.pop('pylib_all', False)

        # include python std lib modules.
        # self.include_pylib = kwargs.pop('pylib', self.include_pylib_all)
        self.include_pylib = kwargs.pop('pylib', self.include_pylib_all)

        self._depgraph = defaultdict(dict)
        self._types = {}
        self._last_caller = None
        # path=None, debug=0, excludes=[], replace_paths=[]

        debug = 5 if self.verbose >= 4 else 0
        mf27.ModuleFinder.__init__(self,
                                   path=syspath,
                                   debug=debug,
                                   # debug=3,
                                   excludes=kwargs.get('excludes', []))

    def add_module(self, fqname):
        if fqname in self.modules:
            return self.modules[fqname]
        self.modules[fqname] = m = Module(fqname)
        return m
    
    def run_script(self, pathname):
        # overridden so we can work directly with .pyc files
        # (the stdlig version hardcodes PY_SOURCE below)
        log.debug("run_script(%r)", pathname)
        self.msg(2, "run_script", pathname)
        with open(pathname, 'rb') as fp:
            stuff = (
                "", 
                "rb", 
                imp.PY_COMPILED if pathname.endswith(".pyc") or pathname.endswith(".pyo") else imp.PY_SOURCE
            )
            self.load_module('__main__', fp, pathname, stuff)

    def import_hook(self, name, caller=None, fromlist=None, level=-1):
        old_last_caller = self._last_caller
        try:
            self._last_caller = caller
            # print "      last_CALLER:", caller, "OLD-lastcaller:", old_last_caller
            return mf27.ModuleFinder.import_hook(self, name, caller, fromlist, level)
        finally:
            self._last_caller = old_last_caller

    def _add_import(self, module):
        if module is not None:
            if self._last_caller:
                # self._depgraph[self._last_caller.__name__][module.__name__] = module.__file__
                if hasattr(module, '__file__') or self.include_pylib_all:
                    pylib_p = []
                    if not module.__file__:
                        pass
                    else:
                        rpath = os.path.split(module.__file__)[0].lower()
                        if 'site-packages' not in rpath:
                            pylib_p = [rpath.startswith(pp) for pp in PYLIB_PATH]
                    if self.include_pylib or not any(pylib_p):
                        # if self._last_caller.__name__ != module.__name__:
                        #     self._depgraph[self._last_caller.__name__][module.__name__] = module.__file__
                        self._depgraph[self._last_caller.__name__][module.__name__] = module.__file__

    def import_module(self, partnam, fqname, parent):
        module = mf27.ModuleFinder.import_module(self, partnam, fqname, parent)
        self._add_import(module)
        return module

    def load_module(self, fqname, fp, pathname, suffix_mode_kind):
        # log.debug("load_module(%r, %r, %r, %r)", fqname, fp, pathname, suffix_mode_kind)
        (suffix, mode, kind) = suffix_mode_kind
        try:
            module = mf27.ModuleFinder.load_module(
                self, 
                fqname, fp, pathname, (suffix, mode, kind)
            )
        except SyntaxError:
            # this happened first when pyinvoke tried to load yaml3/2 based on
            # an `if six.PY3`
            # print "SYNTAX_ERROR"
            module = None
        except AttributeError as e:
            # See issues #139 and #140...
            print("ERROR trying to load {} from {} (py38+ _find_module reimplementation of imp.find_module called .is_package on None...)\n{}".format(
                fqname, pathname,
                e
            ))
            module = None

        if module is not None:
            self._types[module.__name__] = kind
        return module

    def ensure_fromlist(self, module, fromlist, recursive=0):
        self.msg(4, "ensure_fromlist", module, fromlist, recursive)
        for sub in fromlist:
            # print "  for sub:", sub, "in fromlisst:", fromlist, "hasattr(module, sub):", hasattr(module, sub)
            if sub == "*":
                # print "STAR"
                if not recursive:
                    submodules = self.find_all_submodules(module)
                    if submodules:
                        self.ensure_fromlist(module, submodules, 1)
            elif not hasattr(module, sub):
                # print "ELIF......"
                subname = "%s.%s" % (module.__name__, sub)
                submod = self.import_module(sub, subname, module)
                if not submod:
                    raise ImportError("No module named " + subname)
            else:
                self._add_import(getattr(module, sub))
                # print "  SUB:", sub, "lastcaller:", self._last_caller


class RawDependencies(object):
    def __init__(self, fname, **kw):
        path = sys.path[:]
        exclude = []
        mf = MyModuleFinder(path, exclude, **kw)
        mf.run_script(fname)
        self.depgraph = mf._depgraph
        self.types = mf._types


def py2dep(target, **kw) -> depgraph.DepGraph:
    """"Calculate dependencies for ``pattern`` and return a DepGraph.
    """
    log.info("py2dep(%r)", target)
    dummy = DummyModule(target, **kw)

    kw['dummyname'] = dummy.fname
    syspath = sys.path[:]
    syspath.insert(0, target.syspath_dir)

    # remove exclude so we don't pass it twice to modulefinder
    exclude = ['migrations'] + kw.pop('exclude', [])
    log.debug("Exclude: %r", exclude)
    log.debug("KW: %r", kw)
    if 'fname' in kw:
        del kw['fname']

    mf = MyModuleFinder(
        syspath,                # module search path for this module finder
        excludes=exclude,       # folders to exclude
        **kw
    )
    mf.debug = max(mf.debug, kw.get('debug_mf', 0))
    if log.isEnabledFor(logging.DEBUG):
        log.debug("CURDIR: %s", os.getcwd())
        log.debug("FNAME: %r, CONTENT:\n%s\n", dummy.fname, dummy.text())
    mf.run_script(dummy.fname)

    log.info("mf._depgraph:\n%s", json.dumps(dict(mf._depgraph), indent=4))
    log.info("mf.badmodules:\n%s", json.dumps(mf.badmodules, indent=4))

    if kw.get('include_missing'):
        for k, vdict in list(mf.badmodules.items()):
            if k not in mf._depgraph:
                mf._depgraph[k] = {}
            for v in vdict:
                if not target.is_pysource and v not in mf._depgraph['__main__']:
                    mf._depgraph['__main__'][v] = None
                if v in mf._depgraph:
                    mf._depgraph[v][k] = None
                else:
                    mf._depgraph[v] = {k: None}

    log.info("mf._depgraph:\n%s", json.dumps(dict(mf._depgraph), indent=4))

    kw['exclude'] = exclude

    if kw.get('pylib'):
        mf_depgraph = mf._depgraph
        for k, v in list(mf._depgraph.items()):
            log.debug('depgraph item: %r %r', k, v)
        # mf_modules = {k: os.syspath.abspath(v.__file__)
        #               for k, v in mf.modules.items()}
    else:
        pylib = pystdlib()
        mf_depgraph = {}
        for k, v in list(mf._depgraph.items()):
            log.debug('depgraph item: %r %r', k, v)
            if k in pylib:
                continue
            vals = {vk: vv for vk, vv in v.items() if vk not in pylib}
            mf_depgraph[k] = vals

        # mf_modules = {k: os.syspath.abspath(v.__file__)
        #               for k, v in mf.modules.items()
        #               if k not in pylib}

    try:
        import yaml
        log.info("mf_depgraph:\n%s",
                 yaml.dump(dict(mf_depgraph), default_flow_style=False))
        # log.error("mf._types:\n%s", yaml.dump(mf._types, default_flow_style=False))
        # log.debug("mf_modules:\n%s", yaml.dump(mf_modules, default_flow_style=False))
    except ImportError:
        log.info("mf_depgraph:\n%s", json.dumps(dict(mf_depgraph), indent=4))

    return depgraph.DepGraph(mf_depgraph, mf._types, target, **kw)


def py2depgraph():
    _fname = sys.argv[1]
    _graph = RawDependencies(_fname)

    sys.stdout.write(
        json.dumps(_graph.__dict__, indent=4)
    )


if __name__ == '__main__':
    py2depgraph()



# -*- coding: utf-8 -*-
from __future__ import print_function
from tests.filemaker import create_files
from tests.simpledeps import simpledeps, depgrf


def test_no_skip():
    files = """
        relimp:
            - __init__.py
            - a.py: |
                from . import b
            - b.py: |
                from . import c
            - c.py
    """
    with create_files(files) as workdir:
        print("plain", simpledeps('relimp'))
        assert simpledeps('relimp') == {
            'relimp.b -> relimp.a',
            'relimp.c -> relimp.b'
        }


def test_skip_module_pattern():
    files = """
        relimp:
            - __init__.py
            - a.py: |
                from . import b
            - b.py: |
                from . import c
            - c.py
    """
    with create_files(files) as workdir:
        print("-x", simpledeps('relimp', '-x relimp.*'))
        assert simpledeps('relimp', '-x relimp.*') == set()


def test_skip_exact_pattern():
    files = """
        relimp:
            - __init__.py
            - a.py: |
                from . import b
                from . import c
            - b.py: |
                from .c import d
            - c:
                - __init__.py           
                - d.py
    """
    with create_files(files) as workdir:
        print('plain', simpledeps('relimp'))
        assert simpledeps('relimp') == {
            'relimp.b -> relimp.a',
            'relimp.c.d -> relimp.b',
            'relimp.c -> relimp.b',
            'relimp.c -> relimp.a',
        }

        print('plain', simpledeps('relimp', '-xx relimp.c'))
        assert simpledeps('relimp', '-xx relimp.c') == {
            'relimp.b -> relimp.a',
            'relimp.c.d -> relimp.b',  # this should remain
            # 'relimp.c -> relimp.b',  # these should be filtered away..
            # 'relimp.c -> relimp.a',
        }


def test_skip_exact():
    files = """
        relimp:
            - __init__.py
            - a.py: |
                from . import b
            - b.py: |
                from . import c
            - c.py
    """
    with create_files(files) as workdir:
        print('-xx', simpledeps('relimp', '-xx relimp.c'))
        assert simpledeps('relimp', '-xx relimp.c') == {
            'relimp.b -> relimp.a'
        }


def test_skip_modules():
    files = """
        relimp:
            - __init__.py
            - a.py: |
                from . import b
            - b.py: |
                from . import c
            - c.py
    """
    with create_files(files) as workdir:
        assert simpledeps('relimp', '-x relimp.c') == {
            'relimp.b -> relimp.a'
        }
        # g = depgrf('relimp', '-x relimp.c relimp.b')
        # print g
        # print simpledeps('relimp', '-x relimp.c relimp.b')
        # assert 1
        # assert g == 42


def test_rawdeps():
    files = """
        relimp:
            - __init__.py
            - a.py: |
                from . import b
            - b.py: |
                from . import c
            - c.py
    """
    with create_files(files) as workdir:
        assert simpledeps('relimp', '--show-raw-deps -x relimp.c') == {
            'relimp.b -> relimp.a'
        }



# -*- coding: utf-8 -*-
"""
Python module dependency visualization. This package installs the ``pydeps``
command, and normal usage will be to use it from the command line.
"""
__version__ = "1.12.17"



# -*- coding: utf-8 -*-
import os

from pydeps.cli import error
from pydeps.pydeps import pydeps
from tests.filemaker import create_files
from tests.simpledeps import simpledeps, empty


def test_output(tmpdir):
    files = """
        unrelated: []
        foo:
            - __init__.py
            - a.py: |
                from bar import b
        bar:
            - __init__.py
            - b.py
    """
    with create_files(files) as workdir:
        assert os.getcwd() == workdir

        outname = os.path.join('unrelated', 'foo.svg')
        assert not os.path.exists(outname)
        pydeps(fname='foo', **empty('--noshow --show-dot', output=outname))
        assert os.path.exists(outname)


def test_rankdir_default(tmpdir, capsys):
    files = """
        unrelated: []
        foo:
            - __init__.py
            - a.py: |
                from bar import b
        bar:
            - __init__.py
            - b.py
    """
    with create_files(files) as workdir:
        assert os.getcwd() == workdir

        outname = os.path.join('unrelated', 'foo.svg')
        pydeps(fname='foo', **empty('--noshow --show-dot', output=outname))
        captured_stdout = capsys.readouterr().out
        assert 'rankdir = TB' in captured_stdout


def test_error(capsys):
    """Test that error function prints reminder about missing inits on FileNotFoundErrors."""
    try:
        error("[Errno 2] No such file or directory: 'foo'")
    except SystemExit:
        # because error invokes sys.exit(1), we have to catch it here, otherwise the test would always fail.
        pass
    else:  # test should fail if error function doesn't raise
        assert False
    captured_stdout = capsys.readouterr().out
    assert "(Did you forget to include an __init__.py?)" in captured_stdout


def test_rankdir_BT(tmpdir, capsys):
    files = """
        unrelated: []
        foo:
            - __init__.py
            - a.py: |
                from bar import b
        bar:
            - __init__.py
            - b.py
    """
    with create_files(files) as workdir:
        assert os.getcwd() == workdir

        outname = os.path.join('unrelated', 'foo.svg')
        pydeps(fname='foo', **empty('--noshow --show-dot --rankdir=BT', output=outname))
        captured_stdout = capsys.readouterr().out
        assert 'rankdir = BT' in captured_stdout



# -*- coding: utf-8 -*-
from pydeps.pydeps import pydeps
from tests.filemaker import create_files
from tests.simpledeps import simpledeps
import pytest


@pytest.mark.skip(reason="TODO: fix this (issue #174)")
def test_dirtree():
    files = """
        foo:
            - a:
                - __init__.py: ''
                - a.py: |
                    from b.b import bval
            - b:
                - __init__.py: ''
                - b.py: |
                    bval = 42
    """
    with create_files(files) as workdir:
        assert simpledeps('foo', '--show-deps -LINFO -vv') == {
            'b.b -> a.a',
        }



# -*- coding: utf-8 -*-
import json
import os
from pydeps import pydeps
from tests.filemaker import create_files
from tests.simpledeps import simpledeps, depgrf


def test_dep2dot():
    files = """
        foo:
            - __init__.py
            - a.py: |
                from . import b
            - b.py
    """
    with create_files(files) as workdir:
        g = depgrf("foo")
        d = json.loads(repr(g))
        print(d)
        assert '__main__' in d['foo']['imported_by']
        assert g.sources['foo.a'] == g.sources['foo.a']
        assert str(g.sources['foo.a']).startswith('foo.a')
        assert 'foo.b' in repr(g.sources['foo.a'])



# -*- coding: utf-8 -*-
from tests.filemaker import create_files
from tests.simpledeps import simpledeps


def test_from_html5lib():
    files = """
        foo:
            - __init__.py
            - a.py: |
                from bar import py
        bar:
            - __init__.py
    """
    with create_files(files) as workdir:
        assert simpledeps('foo', '--show-deps -LINFO -vv') == {
            'bar -> foo.a',
        }
